""" 
Diematic daemon.

(C) 2022 by Ignacio Hernández-Ros
Based on a previous work from Germain Masse 2019

See and respect licensing terms

"""
import logging
import yaml
import os
import signal
import time
import threading
import argparse
import sys

from lockfile import pidlockfile
from boiler import Boiler 
from pymodbus.client.sync import ModbusSerialClient as ModbusClient
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError
from daemon import DaemonContext
from daemon import pidfile
from aiohttp import web
import asyncio

from webserver import DiematicWebRequestHandler

"""

For InfluxDB backend

influx -precision rfc3339

CREATE DATABASE "diematic"
GRANT ALL ON "diematic" TO "diematic"
CREATE RETENTION POLICY "one_day" ON "diematic" DURATION 24h REPLICATION 1 DEFAULT
CREATE RETENTION POLICY "five_weeks" ON "diematic" DURATION 5w REPLICATION 1
CREATE RETENTION POLICY "five_years" ON "diematic" DURATION 260w REPLICATION 1

CREATE CONTINUOUS QUERY "cq_month" ON "diematic" BEGIN
  SELECT mean(/temperature/) AS "mean_1h", mean(/pressure/) AS "mean_1h", max(/temperature/) AS "max_1h", max(/pressure/) AS "max_1h"
  INTO "five_weeks".:MEASUREMENT
  FROM "one_day"."diematic"
  GROUP BY time(1h),*
END

CREATE CONTINUOUS QUERY "cq_year" ON "diematic" BEGIN
  SELECT mean(/^mean_.*temperature/) AS "mean_24h", mean(/^mean_.*pressure/) AS "mean_24h", max(/^max_.*temperature/) AS "max_24h", max(/^max_.*pressure/) AS "max_24h"
  INTO "five_years".:MEASUREMENT
  FROM "five_weeks"."diematic"
  GROUP BY time(24h),*
END

DROP CONTINUOUS QUERY cq_month ON diematic
DROP CONTINUOUS QUERY cq_year ON diematic

"""

# --------------------------------------------------------------------------- #
# configure the client logging
# --------------------------------------------------------------------------- #
FORMAT = ('%(asctime)-15s %(threadName)-15s '
          '%(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s')
logging.basicConfig(format=FORMAT, level=logging.ERROR)
log = logging.getLogger()

DEFAULT_LOGGING = 'error'

DEFAULT_MODBUS_TIMEOUT = 10
DEFAULT_MODBUS_BAUDRATE = 9600
DEFAULT_MODBUS_UNIT = 10
DEFAULT_MODBUS_DEVICE = None

class DaemonRunnerError(Exception):
    """ Abstract base class for errors from DaemonRunner. """

class DaemonRunnerInvalidActionError(DaemonRunnerError, ValueError):
    """ Raised when specified action for DaemonRunner is invalid. """

class DaemonRunnerStartFailureError(DaemonRunnerError, RuntimeError):
    """ Raised when failure starting DaemonRunner. """


class DaemonRunnerStopFailureError(DaemonRunnerError, RuntimeError):
    """ Raised when failure stopping DaemonRunner. """

class DaemonRunnerReloadFailureError(DaemonRunnerError, RuntimeError):
    """ Raised when failure reloading DaemonRunner. """

class DiematicModbusError(RuntimeError):
    """ Happens when modbus communication is error """

class DiematicApp:
    """ Application container """

    start_message = "started with pid {pid:d}"

    def __init__(self):
        # --------------------------------------------------------------------------- #
        # set configuration variables (command line prevails on configuration file)
        # --------------------------------------------------------------------------- #
        self.MODBUS_TIMEOUT = None
        self.MODBUS_BAUDRATE = None
        self.MODBUS_UNIT = None
        self.MODBUS_DEVICE = None

        self.first_run = True

        self.connection_semaphore = threading.Semaphore()

        return

    def _get_context(self):
        """ Returns or create and return the self.context that is used by the surrounding daemon """
        try:
            return self.context
        except AttributeError:
            self.context = DaemonContext(
                pidfile=pidlockfile.PIDLockFile('/run/diematic/diematicd.pid'),
                working_directory="/etc/diematic"
                )

            self.context.signal_map = {
                signal.SIGTERM: self._terminate_daemon_process,
                signal.SIGHUP: self._terminate_daemon_process,
                signal.SIGUSR1: self._reload_configuration,
                }
            self.context.app = self

            self.pidfile_timeout = 3
            # self._open_streams_from_app_stream_paths()
            return self.context

    # def _get_executor(self):
    #     """ create the executor pool or return if already created """
    #     try:
    #         return self.executor
    #     except AttributeError:
    #         self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
    #         return self.executor

    def _value_writer(self):
        """ consumes a job writing to the boiler """
        write = self.MyBoiler.next_write()
        tryCount = 0
        while not write is None and 'name' in write:
            paramName = write['name']
            winfo = self.MyBoiler.prepare_write(write)
            address = winfo['address']
            newvalue = winfo['value']
            log.info("Pending write {register} address {address} newvalue {newvalue}".format(register=paramName, address=address, newvalue=newvalue))
            if tryCount > 5:
                # too many attemnts to write a value
                self.MyBoiler.write_error(paramName, "write operation failed, too many attempts to write parameter {parameterName} value {wvalue} in address {address}".format(parameterName=paramName, wvalue=newvalue, address=address))
                return
            try:
                self.connection_semaphore.acquire()
                log.info("Connection adquired")
                try:
                    with ModbusClient(method='rtu', port=self.MODBUS_DEVICE, timeout=self.MODBUS_TIMEOUT, baudrate=self.MODBUS_BAUDRATE) as client:
                        client.connect()
                        try:
                            log.info("Going to write")
                            rr = client.write_registers(address, newvalue, unit=self.MODBUS_UNIT)
                            if rr.isError():
                                log.error(rr.message)
                                raise DiematicModbusError(rr.message)
                            rr = client.read_holding_registers(address, unit=self.MODBUS_UNIT)
                            if rr.isError():
                                log.error(rr.message)
                                raise DiematicModbusError(rr.message)
                            receivedvalue = rr.registers[0]
                            if not receivedvalue == newvalue:
                                self.MyBoiler.write_error(paramName, "write operation success, but read value differs write value {wvalue} read value {rvalue}".format(wvalue=newvalue, rvalue=receivedvalue))
                            else:
                                self.MyBoiler.write_ok(paramName)
                        finally:
                            log.info("Closing connection")
                            client.close()
                    write = self.MyBoiler.next_write()
                finally:
                    log.info("Releasing semaphore")
                    self.connection_semaphore.release()
            except DiematicModbusError as error:
                tryCount += 1
                self.MyBoiler.write_error(paramName, "write operation failed, {errormessage}".format(errormessage=error))
                log.info("Repeat in one second")
                time.sleep(1)
                pass

    def run(self):
        self._reload_configuration(None,None)

        if self.args.server == 'both' or self.args.server == 'loop':
            self.start_main_program()

        if self.args.server == 'both' or self.args.server == 'web':
            self.startWebServer()

        if self.args.server == 'both':
            loop = asyncio.get_event_loop()
            loop.run_forever()


    def start_main_program(self):
        x = threading.Thread(target=self.main_program_loop)
        x.setDaemon(True)
        x.start()
        if self.args.server == 'loop':
            x.join()

    def main_program_loop(self):
        while True:
            self.do_main_program()
            time.sleep(60) # a minute

    def startWebServer(self):
        self._create_boiler()

        handler = DiematicWebRequestHandler(self.MyBoiler)

        loop = asyncio.get_event_loop()
        self.webServer = web.Application()
        self.webServer["mainapp"] = self
        self.webServer.add_routes(handler.routes)

        runner = web.AppRunner(self.webServer)
        loop.run_until_complete(runner.setup())
        site = web.TCPSite(runner, self.args.hostname, self.args.port)
        loop.run_until_complete(site.start())
        if self.args.server == 'web':
            loop.run_forever()

    def check_pending_writes(self):
        self._get_executor().submit(self._value_writer)

    def _create_boiler(self):
        if self.shall_create_boiler:
            self.MyBoiler = Boiler(uuid=self.cfg['boiler']['uuid'], index=self.cfg['registers'])
            self.shall_create_boiler = False

    def do_main_program(self):
        self._create_boiler()
        loop = True
        while loop:
            try:
                self.run_sync_client()
                loop = False
            except DiematicModbusError:
                time.sleep(1)
                pass
        return

    def run_sync_client(self):
        #enabling modbus communication
        if self.first_run:
            log.info("Connection parameters: device={port!r} timeout={timeout!r} baudrate={baudrate!r}".format(port=self.MODBUS_DEVICE, timeout=self.MODBUS_TIMEOUT, baudrate=self.MODBUS_BAUDRATE))
            self.first_run = False
        self.connection_semaphore.acquire()
        try:
            with ModbusClient(method='rtu', port=self.MODBUS_DEVICE, timeout=self.MODBUS_TIMEOUT, baudrate=self.MODBUS_BAUDRATE) as client:
                client.connect()
                try:
                    self.MyBoiler.registers = []
                    id_stop = -1

                    for mbrange in self.cfg['modbus']['register_ranges']:
                        id_start = mbrange[0]
                        self.MyBoiler.registers.extend([None] * (id_start-id_stop-1))
                        id_stop = mbrange[1]

                        log.debug("Attempt to read registers from {} to {}".format(id_start, id_stop))
                        rr = client.read_holding_registers(count=(id_stop-id_start+1), address=id_start, unit=self.MODBUS_UNIT)
                        if rr.isError():
                            log.error(rr.message)
                            raise DiematicModbusError(rr.message)
                            # MyBoiler.registers.extend([None] * (id_stop-id_start+1))
                        else:
                            self.MyBoiler.registers.extend(rr.registers)
                finally:
                    client.close()
        finally:
            self.connection_semaphore.release()

        #parsing registers to push data in Object attributes
        self.MyBoiler.browse_registers()
        log.info("Dumping values\n" + self.MyBoiler.dump())


        #pushing data to influxdb
        if self.args.backend and self.args.backend == 'influxdb':
            timestamp = int(time.time() * 1000) #milliseconds
            influx_json_body = [
            {
                "measurement": "diematic",
                "tags": {
                    "host": "raspberrypi",
                },
                "timestamp": timestamp,
                "fields": self.MyBoiler.fetch_data() 
            }
            ]

            influx_client = InfluxDBClient(self.cfg['influxdb']['host'], self.cfg['influxdb']['port'], self.cfg['influxdb']['user'], self.cfg['influxdb']['password'], self.cfg['influxdb']['database'])

            log.debug("Write points: {0}".format(influx_json_body))
            try:
                influx_client.write_points(influx_json_body, time_precision='ms')
            except InfluxDBClientError as e:
                log.error(e)

    def read_config_file(self):
        # --------------------------------------------------------------------------- #
        # retrieve config from diematic.yaml
        # --------------------------------------------------------------------------- #
        if os.path.isabs(self.args.config):
            config_file = self.args.config
        else:
            main_base = os.path.dirname(__file__)
            config_file = os.path.join(main_base, self.args.config)
        if not os.path.exists(config_file):
            errmsg = "Configuration file not found {file!r}".format(file=config_file)
            raise FileNotFoundError(errmsg)
        with open(config_file) as f:
            # use safe_load instead load
            self.cfg = yaml.safe_load(f)

    def toJSON(self):
        """Return the configuration file in json format"""
        return self.cfg["registers"]

    def set_logging_level(self):
        # --------------------------------------------------------------------------- #
        # set logging level
        # --------------------------------------------------------------------------- #
        new_logging_level = DEFAULT_LOGGING
        if self.args.logging:
            new_logging_level = self.args.logging
        elif 'logging' in self.cfg:
            new_logging_level = self.cfg['logging']
        numeric_level = getattr(logging, new_logging_level.upper())
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % new_logging_level)
        log.setLevel(numeric_level)

    def _open_streams_from_app_stream_paths(self):
        """ Open the `daemon_context` streams from the paths specified.

            :param app: The application instance.

            Open the `daemon_context` standard streams (`stdin`,
            `stdout`, `stderr`) as stream objects of the appropriate
            types, from each of the corresponding filesystem paths
            from the `app`.
            """
        self.stdin_path = "/dev/null"
        self.stdout_path = os.path.join(os.sep, "var", "log", "diematic", "diematic.out") # Can also be /dev/null 
        self.stderr_path =  os.path.join(os.sep, "var", "log", "diematic", "diematic.err") # Can also be /dev/null

        self._get_context().stdin = open(self.stdin_path, 'rt')
        self._get_context().stdout = open(self.stdout_path, 'w+t')
        self._get_context().stderr = open(self.stderr_path, 'w+t')

    def _start(self):
        """ Open the daemon context and run the application.

            :return: ``None``.
            :raises DaemonRunnerStartFailureError: If the PID file cannot
                be locked by this process.
            """
        if is_pidfile_stale(self._get_context().pidfile):
            self._get_context().pidfile.break_lock()
        elif self.args.server == 'both' and is_process_already_running(self._get_context().pidfile):
            error = DaemonRunnerStartFailureError(
                "Process already running {pid:d}".format(pid=self._get_context().pidfile.read_pid()))
            raise error

        if self.args.server == 'both' and not self.args.foreground:
            try:
                self._open_streams_from_app_stream_paths()
                self._get_context().open()
            except pidlockfile.AlreadyLocked as exc:
                error = DaemonRunnerStartFailureError(
                        "PID file {pidfile.path!r} already locked".format(
                            pidfile=self._get_context().pidfile))
                raise error from exc

            pid = os.getpid()
            message = self.start_message.format(pid=pid)
            emit_message(message, sys.stdout)

        self.run()
    
    def _runonce(self):
        self._reload_configuration(None,None)
        if self.args.server == 'web':
            self.startWebServer()
        else:
            self.do_main_program()

    def _readregister(self):
        """ Read the content of the indicated register and shows the
            value. Does not need to exist in the yaml file
        """
        address = self.args.address
        format = self.args.format
        self._reload_configuration(None,None)
        self._create_boiler()
        tryCount = 0
        self.connection_semaphore.acquire()
        try:
            while tryCount < 5:
                registers = []
                tryCount += 1
                try:
                    with ModbusClient(method='rtu', port=self.MODBUS_DEVICE, timeout=self.MODBUS_TIMEOUT, baudrate=self.MODBUS_BAUDRATE) as client:
                        client.connect()
                        try:
                            registers.extend([None] * (-1))
                            log.debug("Attempt to read register {}".format(address))
                            rr = client.read_holding_registers(count=1, address=address, unit=self.MODBUS_UNIT)
                            if rr.isError():
                                log.error(rr.message)
                                raise DiematicModbusError(rr.message)
                            else:
                                registers.extend(rr.registers)
                                # format output
                                if format == 'Raw':
                                    emit_message("Register {} value {}".format(address, registers[0]), sys.stdout)
                                elif format == 'DiematicOneDecimal':
                                    emit_message("Register {} value {}".format(address, self.MyBoiler._decode_decimal(registers[0], 1)), sys.stdout)
                                elif format == 'DiematicModeFlag':
                                    emit_message("Register {} value {}".format(address, self.MyBoiler._decode_modeflag(registers[0])), sys.stdout)
                                elif format == 'ErrorCode':
                                    emit_message("Register {} value {}".format(address, self.MyBoiler._decode_errorcode(registers[0])), sys.stdout)
                                elif format == 'DiematicCircType':
                                    emit_message("Register {} value {}".format(address, self.MyBoiler._decode_circtype(registers[0])), sys.stdout)
                                elif format == 'DiematicProgram':
                                    emit_message("Register {} value {}".format(address, self.MyBoiler._decode_program(registers[0])), sys.stdout)
                                elif format == 'bit0':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0001) >> 0), sys.stdout)
                                elif format == 'bit1':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0002) >> 1), sys.stdout)
                                elif format == 'bit2':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0004) >> 2), sys.stdout)
                                elif format == 'bit3':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0008) >> 3), sys.stdout)
                                elif format == 'bit4':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0010) >> 4), sys.stdout)
                                elif format == 'bit5':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0020) >> 5), sys.stdout)
                                elif format == 'bit6':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0040) >> 6), sys.stdout)
                                elif format == 'bit7':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0080) >> 7), sys.stdout)
                                elif format == 'bit8':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0100) >> 8), sys.stdout)
                                elif format == 'bit9':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0200) >> 9), sys.stdout)
                                elif format == 'bitA':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0400) >> 10), sys.stdout)
                                elif format == 'bitB':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x0800) >> 11), sys.stdout)
                                elif format == 'bitC':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x1000) >> 12), sys.stdout)
                                elif format == 'bitD':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x2000) >> 13), sys.stdout)
                                elif format == 'bitE':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x4000) >> 14), sys.stdout)
                                elif format == 'bitF':
                                    emit_message("Register {} value {}".format(address, (registers[0] & 0x8000) >> 15), sys.stdout)
                                tryCount = 99 # exit while
                        finally:
                            client.close()
                except DiematicModbusError:
                    time.sleep(1)
                    pass
        finally:
            self.connection_semaphore.release()        

    def _terminate_daemon_process(self, _signal, _stack):
        """ Terminate the daemon process specified in the current PID file.
            _signal and _stack are required because this function is a signal handler althrough they are not used

            :return: ``None``.
            :raises DaemonRunnerStopFailureError: If terminating the daemon
                fails with an OS error.
            """
        pid = self._get_context().pidfile.read_pid()
        ownpid = os.getpid()
        if pid == ownpid:
            sys.exit(2)
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError as exc:
            error = DaemonRunnerStopFailureError(
                    "Failed to terminate {pid:d}: {exc}".format(
                        pid=pid, exc=exc))
            raise error from exc

    def _stop(self):
        """ Exit the daemon process specified in the current PID file.

            :return: ``None``.
            :raises DaemonRunnerStopFailureError: If the PID file is not
                already locked.
            """
        if not self._get_context().pidfile.is_locked():
            error = DaemonRunnerStopFailureError(
                    "PID file {pidfile.path!r} not locked".format(
                        pidfile=self._get_context().pidfile))
            raise error

        if is_pidfile_stale(self._get_context().pidfile):
            self._get_context().pidfile.break_lock()
        else:
            self._terminate_daemon_process(None, None)

    def _status(self):
        if is_process_already_running(self._get_context().pidfile):
            message = "Process is running on pid {pid:d}".format(pid=self._get_context().pidfile.read_pid())
            emit_message(message, sys.stdout)
        elif is_pidfile_stale(self._get_context().pidfile):
            message = "Process is NOT running but a pid file exists"
            emit_message(message, sys.stdout)
        else:
            message = "Process is NOT running"
            emit_message(message, sys.stdout)

    def _reload(self):
        """ Send a SIGUSR1 to the running process so it is forced to reload configuration file."""
        if not self._get_context().pidfile.is_locked():
            error = DaemonRunnerReloadFailureError(
                "PID file {pidfile.path!r} not locked".format(
                    pidfile=self._get_context().pidfile))
            raise error
        if is_pidfile_stale(self._get_context().pidfile):
            self._get_context().pidfile.break_lock()
        else:
            pid = self._get_context().pidfile.read_pid()
            os.kill(pid, signal.SIGUSR1)

    def _reload_configuration(self, _signal, _stack):
        """ Reload the configuration from the configuration file."""
        if self.args.device:
            self.MODBUS_DEVICE = self.args.device

        self.read_config_file()
        self.set_logging_level()
        if 'modbus' in self.cfg:
            if isinstance(self.cfg['modbus']['timeout'], int):
                self.MODBUS_TIMEOUT = self.cfg['modbus']['timeout']
            if isinstance(self.cfg['modbus']['baudrate'], int):
                self.MODBUS_BAUDRATE = self.cfg['modbus']['baudrate']
            if isinstance(self.cfg['modbus']['unit'], int):
                self.MODBUS_UNIT = self.cfg['modbus']['unit']
            if isinstance(self.cfg['modbus']['device'], str):
                self.MODBUS_DEVICE = self.cfg['modbus']['device']
        if self.MODBUS_TIMEOUT is None:
            self.MODBUS_TIMEOUT = DEFAULT_MODBUS_TIMEOUT
        if self.MODBUS_BAUDRATE is None:
            self.MODBUS_BAUDRATE = DEFAULT_MODBUS_BAUDRATE
        if self.MODBUS_UNIT is None:
            self.MODBUS_UNIT = DEFAULT_MODBUS_UNIT

        # --------------------------------------------------------------------------- #
        # check mandatory configuration variables
        # --------------------------------------------------------------------------- #
        if self.MODBUS_DEVICE is None:
            raise ValueError('Modbus device not set')
        self.shall_create_boiler = True

    def _restart(self):
        """ Stop, then start. """
        self._stop()
        self._start()

    def _test(self):
        """ For testing purposes only, not documented."""
        self._reload_configuration(None,None)
        self._create_boiler()
        testValues = (0x28, 0x8001, 0xc8)
        for testValue in testValues:
            test1 = self.MyBoiler._decode_decimal(testValue, 1)
            test2 = self.MyBoiler._encode_decimal(test1, 1)
            if testValue == test2:
                emit_message('test PASS')
            else:
                emit_message('test FAIL')

    """Action functions dictionary."""
    action_funcs = {
        'start': _start,
        'stop': _stop,
        'restart': _restart,
        'status': _status,
        'reload': _reload,
        'runonce': _runonce,
        'readregister': _readregister
    }

    def _get_action_func(self):
        """ Get the function for the specified action.

            :return: The function object corresponding to the specified
                action.
            :raises DaemonRunnerInvalidActionError: if the action is
               unknown.

            The action is specified by the `action` attribute, which is set
            during `parse_args`.
            """
        try:
            func = self.action_funcs[self.action]
        except KeyError:
            error = DaemonRunnerInvalidActionError(
                    "Unknown action: {action!r}".format(
                        action=self.action))
            raise error
        return func

    def do_action(self):
        """ Perform the requested action.

            :return: ``None``.

            The action is specified by the `action` attribute, which is set
            during `parse_args`.
            """
        func = self._get_action_func()
        try:
            func(self)
        except DaemonRunnerError as e:
            log.error("{action!r} error: {errmsg}".format(action=self.action, errmsg=e))

def ActionType(value):
    try:
        DiematicApp.action_funcs[value]
    except KeyError:
        error = DaemonRunnerInvalidActionError(
                "Unknown action: {action!r}".format(action=value))
        raise error
    return value

def _usage_exit(parser):
    """ Emit a usage message, then exit.

        :param argv: The command-line arguments used to invoke the
            program, as a sequence of strings.
        :return: ``None``.
        """
    emit_message(parser.usage)
    # progname = os.path.basename(argv[0])
    usage_exit_code = 2
    # action_usage = "|".join(self.action_funcs.keys())
    # message = "usage: {progname} {usage}".format(
    #         progname=progname, usage=action_usage)
    # emit_message(message)
    sys.exit(usage_exit_code)

def parse_args(app, argv=None):
    """ Parse command-line arguments.

        :param argv: The command-line arguments used to invoke the
            program, as a sequence of strings.

        :return: ``None``.

        The parser expects the first argument as the program name, the
        second argument as the action to perform.

        If the parser fails to parse the arguments, emit a usage
        message and exit the program.
        """
    if argv is None:
        argv = sys.argv[1:]

    # --------------------------------------------------------------------------- #
    # retrieve command line arguments
    # --------------------------------------------------------------------------- #
    parser = argparse.ArgumentParser()
    parser.add_argument(dest='action', choices=['status', 'start', 'stop', 'restart', 'reload', 'runonce', 'readregister'], default="runonce", help="action to take", type=ActionType)
    parser.add_argument("-b", "--backend", choices=['none', 'influxdb'], default='influxdb', help="select data backend (default is influxdb)")
    parser.add_argument("-d", "--device", help="define modbus device")
    parser.add_argument("-f", "--foreground", help="Run in the foreground do not detach process", action="store_true")
    parser.add_argument("-l", "--logging", choices=['critical', 'error', 'warning', 'info', 'debug'], help="define logging level (default is critical)")
    parser.add_argument("-c", "--config", default='/etc/diematic/diematic.yaml', help="alternate configuration file")
    parser.add_argument("-w", "--hostname", default="0.0.0.0", help="web server host name, defaults to 0.0.0.0")
    parser.add_argument("-p", "--port", default=8080, help="web server port, defaults to 8080", type=int)
    parser.add_argument("-s", "--server", choices=['loop','web','both'], default='both', help="servers to start")
    parser.add_argument("-a", "--address", default=0, help="register address to read whe action is readregister", type=int)
    parser.add_argument("-t", "--format", default='Raw', help="value format to apply for register read, default is Raw", choices=['Raw', 'DiematicOneDecimal', 'DiematicModeFlag', 'ErrorCode', 'DiematicCircType', 'DiematicProgram', 'bit0', 'bit1', 'bit2', 'bit3', 'bit4', 'bit5', 'bit6', 'bit7', 'bit8', 'bit9', 'bitA', 'bitB', 'bitC', 'bitD', 'bitE', 'bitF'])

    if len(argv) < 1:
        _usage_exit(parser, argv)

    app.args = parser.parse_args(argv)

    app.action = app.args.action
    if app.args.action == 'runonce':
        if not ('-b' in argv or 'backend' in argv):
            app.args.backend = 'none'
        if not ('-l' in argv or 'loggin' in argv):
            app.args.logging = 'info'

    if app.args.action == 'readregister':
        app.args.backend = 'none'

    # self.action = str(argv[1])
    if app.action not in app.action_funcs:
        _usage_exit(parser, argv)

def emit_message(message, stream=None):
    """ Emit a message to the specified stream (default `sys.stderr`). """
    if stream is None:
        stream = sys.stderr
    stream.write("{message}\n".format(message=message))
    stream.flush()

def make_pidlockfile(path, acquire_timeout):
    """ Make a PIDLockFile instance with the given filesystem path. """
    if not isinstance(path, str):
        error = ValueError("Not a filesystem path: {path!r}".format(
                path=path))
        raise error
    if not os.path.isabs(path):
        error = ValueError("Not an absolute path: {path!r}".format(
                path=path))
        raise error
    lockfile = pidfile.TimeoutPIDLockFile(path, acquire_timeout)

    return lockfile

def is_pidfile_stale(pidfile):
    """ Determine whether a PID file is stale.

        :return: ``True`` if the PID file is stale; otherwise ``False``.

        The PID file is “stale” if its contents are valid but do not
        match the PID of a currently-running process.
        """
    result = False

    pidfile_pid = pidfile.read_pid()
    if pidfile_pid is not None:
        try:
            os.kill(pidfile_pid, signal.SIG_DFL)
        except ProcessLookupError:
            # The specified PID does not exist.
            result = True

    return result

def is_process_already_running(pidfile):
    """ Determine if the process indicated by the pidfile is already
        running.
        :return: ``True`` if the process pointed to by the pid file is running
        """
    result = False
    pidfile_pid = pidfile.read_pid()
    if pidfile_pid is not None:
        try:
            os.kill(pidfile_pid, signal.SIG_DFL)
            result = True
        except ProcessLookupError:
            # The specified PID does not exist.
            pass

    return result

if __name__ == '__main__':
    app = DiematicApp()
    parse_args(app)
    app.do_action()



import logging

from datetime import datetime

log = logging.getLogger()

class Boiler:
    """ Class representation of a De Dietrich boiler with capacity to read registers
        :param index: instance of the yaml configuration file
    """
    def __init__(self, uuid, index):
        self.uuid = uuid
        self.registers = []
        self.attribute_list = []
        self.index = index
        for register in self.index:
            influx = True
            if 'influx' in register:
                influx = register['influx']

            if register['type'] == 'bits':
                for varname in register['bits']:
                    self._init_register_value(varname, register['id'], influx)
                    self.attribute_list.append(varname)
            else:
                self._init_register_value(register['name'], register['id'], influx)
                self.attribute_list.append(register['name'])

    def _init_register_value(self, varname, id, influx):
        setattr(self, varname, {'name': varname, 'status': 'init', 'value': None, 'id': id, 'influx': influx})

#       {
#           'name': varname,
#           'value': registervalue, 
#           'read': datetime.now().isoformat(), 
#           'status': can be 'init', 'read', 'writepending' or 'checking' or 'error'
#           'newvalue': this is the new value to be written, checked etc. status goes to 'read' when 'newvalue' exist and is equal to 'registervalue'
#           'error': error message after writing a value and a failure received
#       }
    def _set_register_value(self, varname, registervalue):
        previousValue = getattr(self, varname, {'name': varname, 'status': 'init', 'value': registervalue})
        varvalue = previousValue.copy()
        prestatus = previousValue.get('status')
        if prestatus == 'init':
            status = 'read'
        else:
            status = prestatus
        if status != 'init':
            varvalue['value'] = registervalue
            varvalue['read'] = datetime.now().isoformat()
        varvalue['status'] = status
        setattr(self, varname, varvalue)

    def _decode_decimal(self, value_int, decimals=0):
        if (value_int == 65535):
            return None
        else:
            output = value_int & 0x7FFF
        if (value_int >> 15 == 1):
            output = -output
        return float(output)/10**decimals

    def _encode_decimal(self, value, decimals=0):
        decimalvalue = int(value*10**decimals)
        if decimalvalue < 0:
            return (-decimalvalue & 0x7FFF) | 0x8000
        return decimalvalue & 0x7FFF

    def _decode_errorcode(self, value_int):
        if (value_int == 0x0000):
            return 'OK'
        if (value_int == 0x0001):
            return 'BOILER S.FAIL.'
        if (value_int == 0x0002):
            return 'OUTL S.A FAIL.'
        if (value_int == 0x0003):
            return 'OUTL S.B FAIL.'
        if (value_int == 0x0004):
            return 'OUTL S.C FAIL.'
        if (value_int == 0x0005):
            return 'OUTSI. S.FAIL.'
        if (value_int == 0x0006):
            return 'SMOKE S. FAIL.'
        if (value_int == 0x0007):
            return 'AUX. F. DEFEKT'
        if (value_int == 0x0009):
            return 'DHW S. FAILURE'
        if (value_int == 0x000A):
            return 'BACK S.FAILURE'
        if (value_int == 0x000B):
            return 'ROOM S.A FAIL.'
        if (value_int == 0x000C):
            return 'ROOM S.B FAIL.'
        if (value_int == 0x000D):
            return 'ROOM S.C FAIL.'
        if (value_int == 0x000E):
            return 'SOLAR S. FAIL'
        if (value_int == 0x000F):
            return 'ST.TANK S.FAIL'
        if (value_int == 0x0010):
            return 'SWIM.P.A S.FAIL'
        if (value_int == 0x0011):
            return 'DHW 2 S. FAIL'
        if (value_int == 0x0012):
            return 'CDI.A COM.FAIL'
        if (value_int == 0x0013):
            return 'CDI.B COM.FAIL'
        if (value_int == 0x0014):
            return 'CDI.C COM.FAIL'
        if (value_int == 0x001B):
            return 'I-CURRENT FAIL'
        if (value_int == 0x001C):
            return 'BURNER FAILURE'
        if (value_int == 0x001D):
            return 'PARASIT FLAME'
        if (value_int == 0x001E):
            return 'STB BOILER'
        if (value_int == 0x001F):
            return 'STB BACK'
        if (value_int == 0x0020):
            return 'VALVE FAIL'
        if (value_int == 0x0022):
            return 'PCU BLOCKING'
        if (value_int == 0x0023):
            return 'EXCHAN.S.FAIL'
        if (value_int == 0x0024):
            return 'STB EXCHANGE'
        if (value_int == 0x0025):
            return 'TA-S SHORT-CIR'
        if (value_int == 0x0026):
            return 'TA-S DISCONNEC'
        if (value_int == 0x0027):
            return 'TA-S FAILURE'
        if (value_int == 0x0028):
            return 'MC COM.FAIL'
        if (value_int == 0x0029):
            return 'AUX2.SENS.FAIL'
        if (value_int == 0x002A):
            return 'UNIV.SENS.FAIL'
        if (value_int == 0x002B):
            return 'SWIM.P.B S.FAIL'
        if (value_int == 0x002C):
            return 'SWIM.P.C S.FAIL'
        if (value_int == 0x002D):
            return 'PCU COM. FAIL'
        if (value_int == 0x002E):
            return 'LOCKING'
        if (value_int == 0x002F):
            return 'PSU FAIL'
        if (value_int == 0x0030):
            return 'PSU PARAM FAIL'
        if (value_int == 0x0031):
            return 'CCE TEST FAIL'
        if (value_int == 0x0032):
            return 'FAN FAILURE'
        if (value_int == 0x0033):
            return 'SMOKE.P.FAIL'
        if (value_int == 0x0034):
            return 'SU COM.FAIL'
        if (value_int == 0x0035):
            return 'PCU-M3 COM.FAIL'
        if (value_int == 0x0036):
            return 'CS OPEN FAIL'
        if (value_int == 0x0037):
            return 'EXCH-BACK<MIN'
        if (value_int == 0x0038):
            return 'EXCH-BACK>MAX'
        if (value_int == 0x0039):
            return 'BACK>BOIL FAIL'
        if (value_int == 0x003A):
            return 'FAIL UNKNOWN'
        return "Unknown error 0x{errno:x}".format(errno=value_int)

    def _decode_modeflag(self, value_int):
       """ Decodes and normalizes the working mode of the boiler.
            0 -> Anti-freeze
            2 -> Night
            4 -> Day
       """ 
       if value_int not in (0, 2, 4):
           return None
       if value_int == 4:
           return 1
       if value_int == 2:
           return 0
       if value_int == 0:
           return -1

    def _encode_modeflag(self, value):
        if value not in (1,0,-1):
            return None
        if value == 1:
            return 4
        if value == 0:
            return 2
        if value == -1:
            return 0

    def _decode_circtype(self, value_int):
        """ Decodes and normalizes the circuit type mode of the boiler.
            0 -> Disable
            1 -> Direct
            2 -> 3 Way Valve
            3 -> Direct+
            4 -> 3 Way Valve+
            5 -> Swimingpool
        """ 
        if value_int == 0:
           return 'DISABLE'
        if value_int == 1:
            return 'DIRECT'
        if value_int == 2:
            return '3WV'
        if value_int == 3:
            return 'DIRECT+'
        if value_int == 4:
            return '3WV+'
        if value_int == 5:
            return 'SWIM.'
        return 'UNKOWN'

    def _encode_circtype(self, value):
        if value == 'DISABLE':
            return 0
        if value == 'DIRECT':
            return 1
        if value == '3WV':
            return 2
        if value == 'DIRECT+':
            return 3
        if value == '3WV+':
            return 4
        if value == 'SWIM.':
            return 5
        return None

    def _decode_program(self, value_int):
        """ Decodes program applied to circuit.
            0 -> P1
            1 -> P2
            2 -> P3
            3 -> P4
        """ 
        return value_int + 1

    def _encode_program(self, value):
        return value - 1

    def _register(self, varname):
        for register in self.index:
            if not isinstance(register['id'], int):
                return
            if register['type'] == 'bits':
                for i in range(len(register['bits'])):
                    bit_varname = register['bits'][i]
                    if bit_varname == varname:
                        return register
            else:
                if register.get('name') == varname:
                    return register
        return None

    def _update_register(self, register):
        if not isinstance(register['id'], int):
            return
        register_value = self.registers[register['id']]
        if register_value is None:
            log.debug('Browsing register id {:d} value: None'.format(register['id']))
            return
        log.info('Browsing register id {:d} value: {:#04x}'.format(register['id'], register_value))
        if register['type'] == 'bits':
            for i in range(len(register['bits'])):
                bit_varname = register['bits'][i]
                if bit_varname == 'io_unused':
                    continue
                bit_value = register_value >> i & 1
                self._set_register_value(bit_varname, bit_value)
        else:
            varname = register.get('name')
            if varname and varname.strip(): #test name exists
                if register['type'] == 'DiematicOneDecimal':
                    self._set_register_value(varname, self._decode_decimal(register_value, 1))
                elif register['type'] == 'DiematicModeFlag':
                    self._set_register_value(varname, self._decode_modeflag(register_value))
                elif register['type'] == 'ErrorCode':
                    self._set_register_value(varname, self._decode_errorcode(register_value))
                elif register['type'] == 'DiematicCircType':
                    self._set_register_value(varname, self._decode_circtype(register_value))
                elif register['type'] == 'DiematicProgram':
                    self._set_register_value(varname, self._decode_program(register_value))
                else:
                    self._set_register_value(varname, register_value)

    def browse_registers(self):
        for register in self.index:
            self._update_register(register)

    def dump_registers(self):
        output = ''
        for id in range(len(self.registers)):
            if self.registers[id] is None:
                output += "{:d}: None\n".format(id)
            else:
                output += "{:d}: {:#04x}\n".format(id, self.registers[id])
        return output

    def fetch_data(self):
        output = { }
        output['uuid'] = self.uuid
        for varname in self.attribute_list:
            register = getattr(self, varname)
            if register['influx']:
                output[varname] = register['value']
        return output

    def dump(self):
        output = ''
        for varname,value in self.fetch_data().items():
            output += varname + ' = ' + str(value) + "\n"
        return output

    def toJSON(self):
        return self.fetch_data()

    def set_write_pending(self, varname, newvalue):
        value = getattr(self, varname, None)
        if value is None:
            return
        value['newvalue'] = newvalue
        value['status'] = 'writepending'
        setattr(self, varname, value)

    def next_write(self):
        """ returns the next register that contains a pending write or None
        """
        for varname in self.attribute_list:
            value = getattr(self, varname, {})
            if 'status' in value and value['status'] == 'writepending':
                value['status'] = 'checking'
                return value
        return None

    def prepare_write(self, write):
        """ returns a dictionary with two keys:
            the 'address' key contains the register address to write to,
            the 'value' key contains the new value
        """
        register = self._register(write['name'])
        encodedValue = write['newvalue']
        if register['type'] == 'bits':
            overallvalue = 0
            for i in range(len(register['bits'])):
                bit_varname = register['bits'][i]
                if bit_varname == 'io_unused':
                    continue
                if bit_varname != write['name']:
                    bit_value = getattr(self, bit_varname)['value'] >> i
                    overallvalue = overallvalue | bit_value
                else:
                    bit_value = write['newvalue'] & 1 >> i
                    overallvalue = overallvalue | bit_value
            encodedValue = overallvalue
        if register['type'] == 'DiematicOneDecimal':
            encodedValue = self._encode_decimal(encodedValue, 1)
        elif register['type'] == 'DiematicModeFlag':
            encodedValue = self._encode_modeflag(encodedValue)
        elif register['type'] == 'ErrorCode':
            raise ValueError('Cannot write read only value')
        elif register['type'] == 'DiematicCircType':
            encodedValue = self._encode_circtype(encodedValue)
        elif register['type'] == 'DiematicProgram':
            encodedValue = self._encode_program(encodedValue)

        return {
            "address": register['id'],
            "value": encodedValue
        }

    def write_error(self, varname, message):
        """ The write operation failed to compare values """
        value = getattr(self, varname, {})
        value['error'] = message
        value['status'] = 'error'

    def clear_error(self, varname):
        """ clear error on varname """
        value = getattr(self, varname, {})
        value.pop('error', None)
        value.pop('newvalue', None)
        value['status'] = 'read'

    def write_ok(self, varname):
        """ write operation succeed """
        value = getattr(self, varname, {})
        newvalue = value.pop('newvalue')
        value.pop('error', None)
        value['value'] = newvalue
        value['status'] = 'read'



"""
Package definition for the diematic boiler.
We are exposing only the Boiler class. Note that, in order to
create an instance of the Boiler class it is required to read
the configuration file in yaml format that will provide the
registers available in your boiler. There is more information
and examples in the diematicd.py file
"""
from boiler import Boiler


# coding=UTF-8
"""Setup for the pipy package"""
import setuptools

with open('README.md', 'r', encoding='utf-8') as long_description_f:
	long_description = long_description_f.read()

setuptools.setup(
  name = 'diematic_server',
  version = '2.0',
  description = 'Unix daemon and supporting models for publishing data from Diematic DeDietrich boiler',
	long_description = long_description,
	long_description_content_type = 'text/markdown; charset=UTF-8',
  author = 'Ignacio Hernández-Ros',
  author_email = 'ignacio@hernandez-ros.com',
  packages = ['diematic_server'],
  license='MIT',
  url = 'https://github.com/IgnacioHR/diematic_server',
  download_url = 'https://github.com/IgnacioHR/diematic_server/archive/refs/tags/v2.0.tar.gz',
  keywords = ['python', 'home-automation', 'iot', 'influxdb', 'restful', 'modbus', 'de-dietrich', 'diematic'],
  install_requires=[
					'daemon==1.2',
					'influxdb==5.2.3',
					'pymodbus==2.2.0',
					'python-daemon==2.3.0',
					'PyYAML==5.4',
					'aiohttp==3.8.1',
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
		'Environment :: No Input/Output (Daemon)',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
		'Operating System :: Unix',
    'Programming Language :: Python :: 3.7',
		'Topic :: Home Automation',
	],
)


""" 
We will use aiohttp for the new web server. 
This shall provide http 1.1 support and hopefully shall not produce 
aiohttp.client_exceptions.ClientPayloadError: Response payload is not completed
problems on the client side. Let's see how it works!
"""

import json
from aiohttp import web

def _parameter_names(boiler) -> list:
	parameter_names = []
	for register in boiler.index:
		if register['type'] == 'bits':
			for bit in register['bits']:
				if bit != "io_unused":
					parameter_names.append(bit)
		else:
			parameter_names.append(register['name'])
	parameter_names.sort()
	return parameter_names


class DiematicWebRequestHandler:
	""" 
		This class implements the web server that provides GET and POST
		requests for the parameters of the boiler

		The parameters are defined in the same diematic.yaml file

		URL format:
		GET http://{host:port}/diematic/parameters
		returns a list of known parameters from the diematic.yaml

		GET http://{host:port}/diematic/parameters/{parameterName}
		return a JSON 
		{
			"name": "parameterName",
			"status": "read",
			"value": 34,
			"id": 680,
			"influx": true,
			"read": "2022-04-02T17:21:32.751479"
		}

		name: is the parameter name
		status: can be:
			"init": the value has not been read, the record is initialized
			"read": the value has been read
			"writepending": there is a new value pending to be written
			"checking": the value has been written, the boiler is pending reading to check if the new value has been successfully set
			"error": a problem occurred while setting the value
		value: the parameter value
		read: the last time the value was set
		newvalue: when status is "writepending" this record holds the value to be written
		error: the error message when status is "error"

		POST http://{host}/diematic/parameters/{parameterName}
		body must contain a json
		{
			"parameterName": value
		}
	"""

	routes = web.RouteTableDef()
	parameter_names = []

	def __init__(self, boiler) -> None:
		DiematicWebRequestHandler.parameter_names.clear()
		for register in boiler.index:
			if register['type'] == 'bits':
				for bit in register['bits']:
					if bit != "io_unused":
						DiematicWebRequestHandler.parameter_names.append(bit)
			else:
				DiematicWebRequestHandler.parameter_names.append(register['name'])
		DiematicWebRequestHandler.parameter_names.sort()

	@routes.get('/diematic/parameters')
	async def send_list(request):
		""" produces a list of well known register names."""
		scheme = request.scheme
		host = request.host
		document = f"""
<html>
	<head>
		<title>Diematic REST controller by IHR at home (Ignacio Hernández-Ros)</title>
		<style type="text/css">
			body {{
				font-size: 0.9em;
				font-family: sans-serif;
			}}
			.styled-table {{
				border-collapse: collapse;
				margin: 25px 0;
				min-width: 400px;
				box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
			}}
			.styled-table thead tr {{
				background-color: #009879;
				color: #ffffff;
				text-align: left;
			}}
			.styled-table th,
			.styled-table td {{
					padding: 12px 15px;
			}}
			.styled-table tbody tr {{
					border-bottom: 1px solid #dddddd;
			}}
			.styled-table tbody tr:nth-of-type(even) {{
					background-color: #f3f3f3;
			}}
			.styled-table tbody tr:last-of-type {{
					border-bottom: 2px solid #009879;
			}}
		</style>
	</head>
<body>
	<table class="styled-table">
		<thead>
			<tr><th colspan="3">Usage:</th></tr>
		</thead>
		<tbody>
			<tr><td>GET</td><td>{scheme}://{host}/diematic/parameters</td><td><b>returns this page</b></td></tr>
			<tr><td>GET</td><td>{scheme}://{host}/diematic/parameter/{{name}}</td><td><b>returns json with parameter information</b></td></tr>
			<tr><td>GET</td><td>{scheme}://{host}/diematic/config</td><td><b>returns boiler configuration parameters</b></td></tr>
			<tr><td>GET</td><td>{scheme}://{host}/diematic/json</td><td><b>returns all boiler parameters in single json</b></td></tr>
			<tr><td>POST</td><td>{scheme}://{host}/diematic/parameter/{{name}}</td><td><b>Set a parameter value. The body must be a json of this shape {{"value": XX}}. After the POST, the parameters may take some time to be written to the boiler. Use GET with the parameter name for information about the write operation status.</b></td></tr>
			<tr><td>POST</td><td>{scheme}://{host}/diematic/parameter/{{name}}/resume</td><td><b>If, for any reason, a write operation fails, a post like this will reset parameter to normal status.</b></td></tr>
		</tbody>
	</table>
	<p>Recognized parameters list</p>
	<ul>"""
		for name in _parameter_names(request.app["mainapp"].MyBoiler):
			document = document + f"<li><a href='/diematic/parameters/{name}'>{name}</a></li>\n"
		document = document + """</ul></body></html>"""
		return web.Response(text=document, content_type='text/html')

	@routes.get('/diematic/parameters/{paramName}')
	async def send_param(request):
		param_name = request.match_info.get('paramName')
		if not param_name in DiematicWebRequestHandler.parameter_names:
			return web.Response(status=422, reason=f'\'{param_name}\' is an invalid parameter')
		boiler = request.app["mainapp"].MyBoiler
		value = getattr(boiler, param_name)
		return web.json_response(value)

	@routes.post('/diematic/parameters/{paramName}')
	async def set_param(request):
		param_name = request.match_info.get('paramName')
		if not param_name in DiematicWebRequestHandler.parameter_names:
			return web.Response(status=422, reason=f'\'{param_name}\' is an invalid parameter')
		content_len = request.content_length
		if not content_len is None:
			data = await request.content.read(content_len)
		else:
			data = await request.content.read()
		jsoninput = json.loads(data.decode('utf8'))
		value = jsoninput['value']
		mainapp = request.app["mainapp"]
		mainapp.MyBoiler.set_write_pending(param_name, value)
		return web.Response()

	@routes.post('/diematic/parameters/{paramName}/resume')
	async def set_param(request):
		param_name = request.match_info.get('paramName')
		if not param_name in DiematicWebRequestHandler.parameter_names:
			return web.Response(status=422, reason=f'\'{param_name}\' is an invalid parameter')
		mainapp = request.app["mainapp"]
		mainapp.MyBoiler.clear_error(param_name)
		return web.Response()

	@routes.get('/diematic/json')
	async def send_json(request):
		config = request.app["mainapp"].MyBoiler.toJSON()
		return web.json_response(config)

	@routes.get('/diematic/config')
	async def send_config(request):
		config = request.app["mainapp"].toJSON()
		return web.json_response(config)